package PIANOS.datastructures;

import PIANOS.exceptions.*;
import PIANOS.datastructures.*;

import java.util.*;

/*
Note: The thing calling generaterFreqCode and generateGenCode
must add _dp to real constants. For example, "1.0_dp", not "1.0".
*/

public class ContinuousUniformDistribution extends Distribution
{
	public ContinuousUniformDistribution()
    {
	super(2);
	super.parameterType[0] = 1; // real
	super.parameterType[1] = 1; // real
    }

    public ArrayList<String> getIntroduction()
	{
		ArrayList<String> result = new ArrayList<String>();
		result.add("EXTERNAL G05FAF");
        return result;
	}

	public ArrayList<String> getFreqCode(String[] parameters) throws IllegalParametersException
	{
		// This is not done with NAG because it's a trivial calculation
		// Parameters are: m, n, point, result

		if (parameters.length != 4)
		{
			throw new IllegalParametersException("Trying to use continuous uniform distribution with illegal parameters.");
		}

		/*
		The code we want to generate looks like this:
		IF (point < m .OR. point > n) THEN
			result = 0
		ELSE
			result = 1 / (n - (m))
		END IF
		*/

		ArrayList<String> result = new ArrayList<String>();

		result.add("IF (" + parameters[2] + " < " + parameters[0] + " .OR. " +
			parameters[2] + " > " + parameters[1] + ") THEN");
		result.add(parameters[3] + " = 0");
		result.add("ELSE");
		result.add(parameters[3] + " = 1.0 / (" + parameters[1] + " - (" + parameters[0] + "))");
		result.add("END IF");

		return result;
	}

	public ArrayList<String> getGenCode(String[] parameters) throws IllegalParametersException
	{
        // Parameters are: m, n, result

        if (parameters.length != 3)
		{
			throw new IllegalParametersException("Trying to use continuous uniform distribution with illegal parameters.");
		}

        /*
        The code we want to generate looks like this:
        CALL G05FAF(m, n, SIZE(result), result)
        */

        ArrayList<String> result = new ArrayList<String>();

		result.add("CALL G05FAF(" + parameters[0] + ", " + parameters[1] +
            ", SIZE(" + parameters[2] + "), " + parameters[2] + ")");

		return result;
	}
}
