package PIANOS.datastructures;

import PIANOS.exceptions.*;
import PIANOS.datastructures.*;

import java.util.*;

/*
Note: The thing calling generaterFreqCode and generateGenCode
must add _dp to real constants. For example, "1.0_dp", not "1.0".
*/

public class PoissonDistribution extends Distribution {
    public PoissonDistribution() {
        super(1);
        parameterType[0] = 1; // real
    }
    
    public ArrayList<String> getIntroduction() {
        ArrayList<String> result = new ArrayList<String>();
        result.add("EXTERNAL G01BKF");
        result.add("EXTERNAL G05ECF");
        result.add("INTEGER G05EYF");
        result.add("EXTERNAL G05EYF");
        return result;
    }

    public ArrayList<String> getFreqCode(String[] parameters) throws IllegalParametersException {
        // Paramters are: mu, point, result
        
        if (parameters.length != 3) {
            throw new IllegalParametersException("Trying to use poisson distribution with illegal parameters.");
        }
        
        /*
          The code we want to generate looks like this:
          ifail=0
          CALL G01BKF(mu, point, temp_real, temp_real, result, ifail)
          IF (ifail /= 0) THEN
          WRITE (*, *) 'Error when trying to calculate frequency of the poisson distribution, parameters: mu = ', mu
          STOP
          END IF
        */

        // BJF? I hope that was only a TYPO....
        ArrayList<String> result = new ArrayList<String>();
        
        result.add("ifail=0");
        result.add("IF ("+ parameters[0] + " > 1.0D+6) THEN");
        result.add("CALL G01BKF( 1.0D+6, " + parameters[1] +
                   ", temp_real, temp_real, " + parameters[2] + ", ifail)");
        result.add("ELSE");
        result.add("CALL G01BKF(" + parameters[0] + ", " + parameters[1] +
                   ", temp_real, temp_real, " + parameters[2] + ", ifail)");
        result.add("END IF");
        result.add("IF (ifail /= 0) THEN");
        result.add("WRITE (*, *) 'Error when trying to calculate frequency of the poisson distribution, " +
                   "parameters: mu = ', " + parameters[0]);
        result.add("STOP");
        result.add("END IF");
        
        return result;
        
        /*
          Note:
          The code generated by this method uses the following already introduced variables:
        INTEGER :: ifail
        REAL :: temp_real
        */
    }
    
    public ArrayList<String> getGenCode(String[] parameters) throws IllegalParametersException {
        // Parameters are: mu, result
        
        if (parameters.length != 2) {
            throw new IllegalParametersException("Trying to use poisson distribution with illegal parameters.");
        }
        
        /*
          The code we want to generate looks like this:
          ifail=0
          CALL G05ECF(mu, poisson_reference, SIZE(poisson_reference), ifail)
          IF (ifail /= 0) THEN
          WRITE (*, *) 'Error when trying to generate random numbers from poisson distribution, parameters: mu = ', mu
          STOP
          END IF
          DO k=0, SIZE(result)
          result(k)=G05EYF(poisson_reference, SIZE(poisson_reference))
          END DO
        */
        
        ArrayList<String> result = new ArrayList<String>();
        
        result.add("ifail = 0");
        result.add("ALLOCATE(poisson_reference(1 : NINT(20 + 20 * SQRT(" + parameters[0] + "))))");
        /*
          optimal value: 20 +20 sqrt(T)
        */
        result.add("CALL G05ECF(" + parameters[0] + ", poisson_reference, SIZE(poisson_reference), ifail)");
        result.add("IF (ifail /= 0) THEN");
        result.add("WRITE (*, *) 'Error when trying to generate random numbers from poisson distribution, parameters: mu = ', "
                   + parameters[0]);
        result.add("STOP");
        result.add("END IF");
        result.add("DO k=1, SIZE(" + parameters[1] +")");
        result.add(parameters[1] + "(k)=G05EYF(poisson_reference, SIZE(poisson_reference))");
        result.add("END DO");
        
        return result;
        
        /*
          Note:
          The code generated by this method uses the following already introduced variables:
          INTEGER :: k, ifail
          REAL(KIND=dp), DIMENSION(:), ALLOCATABLE :: poisson_reference 
          
          
        */
    }
}
