package PIANOS.generator;
import PIANOS.io.*;
import PIANOS.datastructures.*;
import PIANOS.exceptions.*;
import PIANOS.*;
import java.util.*;
import java.io.*;

public class FortranMain {
    public static void generateMain(String callParameters, ComputationalModel model)
    throws IOException, InvalidModelException, IllegalParametersException,
            MissingFunctionException {
        
        ArrayList<String> result = new ArrayList<String>();
        LinkedList<Variable> variables = model.getVariableList();
        LinkedList<Entity> entities = model.getEntityList();
        
        // Set the topologicalList in Acceptation
        
        ArrayList<Variable> allVariables = new ArrayList<Variable>(variables);
        
        for (Entity entity : entities) {
            allVariables.addAll(entity.getVariableList());
        }
        
        Acceptation.setTopologicalList(allVariables);
        
        
        // NOTE: callParameters are now set in a separate method before
        // any generation. That is because they are needed in
        // other modules in addition to Main.
        
        result.add("PROGRAM main");
        result.add("USE definitions");
        result.add("USE input");
        result.add("USE output");
        result.add("USE proposal");
        result.add("USE user_dist");
        result.add("IMPLICIT NONE ");
        result.add("");
        
        
        //the variables in the model and their type (real or integer)
        for (Variable variableIterator : variables){
            if (variableIterator.isInteger())
                result.add("TYPE(variables_int) :: " + variableIterator.getName());
            else
                result.add("TYPE(variables_real) :: " + variableIterator.getName());
        }
        
        for (Entity entityIterator : entities){
            variables = entityIterator.getVariableList();
            for (Variable variableIterator : variables){
                if (variableIterator.isInteger())
                    result.add("TYPE(variables_int) :: " + variableIterator.getName());
                else
                    result.add("TYPE(variables_real) :: " + variableIterator.getName());
            }
        }
        result.add("");
        
        
        result.add("CHARACTER(LEN=*), PARAMETER :: output_filename='" + model.getOutputFile() +
                "', summary_filename='" + model.getSummaryFile() + "'");
        
        result.add("INTEGER :: i");
        result.add("INTEGER, PARAMETER :: multiplier=100");
        result.add("INTEGER, PARAMETER :: thinning=" + model.getThinning() + ", burn_in=" +
                model.getBurnIn());
        result.add("INTEGER :: status, progress");
        
        Entity spatial = null;
        
        for (int i = 0; i < entities.size(); ++i){
            if (entities.get(i).isSpatial()){
                spatial = entities.get(i);
                break;
            }
        }
        if (spatial != null){
        int size = spatial.getSize();
        
        result.add("INTEGER, DIMENSION("+ size +", "+ (model.getNeighbourCount() +1) +
                ") :: spatial");
            
        }
        result.add("CALL random_init()");
        
        variables = model.getVariableList();
        for (Variable variableIterator : variables){
            result.add("ALLOCATE(" + variableIterator.getName() + " % one_dim(1:1))");
            if(!variableIterator.isFunctional()){
                result.add("ALLOCATE(" + variableIterator.getName() + " % buffer(multiplier))");
                result.add(variableIterator.getName() + " % one_dim  % successful_changes = 0");
                result.add(variableIterator.getName() + " % buffer_index = -1");
            }
        }
        
        for (Entity entityIterator : entities){
            variables = entityIterator.getVariableList();
            
            if (!entityIterator.isMatrix()){
                for (Variable variableIterator : variables){
                    result.add("ALLOCATE(" + variableIterator.getName() + " % one_dim(1:" +
                            entityIterator.getSize() + "))");
                    
                    if (variableIterator.isFunctional())
                        continue;
                    if (variableIterator.isData() && variableIterator.getMissingValueCount() == 0)
                        continue;
                    if (variableIterator.isData() && variableIterator.getMissingValueCount() > 0){
                        
                    result.add("ALLOCATE(" + variableIterator.getName() + 
                            " % one_dim_missing("+ variableIterator.getMissingValueCount() +"))");
                    }
                    
                    result.add("ALLOCATE(" + variableIterator.getName() + "% buffer(multiplier * "
                            + entityIterator.getSize() + "))");
                    result.add(variableIterator.getName() + " % one_dim  % successful_changes = 0");
                    result.add(variableIterator.getName() + " % buffer_index = -1");
                }
                
            }else{
                for (Variable variableIterator : variables){
                    result.add("ALLOCATE(" + variableIterator.getName() + " % two_dim(1:" +
                            entityIterator.getYCoordinate().getSize()
                            + ", 1:" + entityIterator.getXCoordinate().getSize() + "))");
                    
                    if (variableIterator.isFunctional())
                        continue;
                    if (variableIterator.isData() && variableIterator.getMissingValueCount() == 0)
                        continue;
                    if (variableIterator.isData() && variableIterator.getMissingValueCount() > 0){
                    result.add("ALLOCATE(" + variableIterator.getName() + 
                            " % two_dim_missing("+ variableIterator.getMissingValueCount() +", 1:2))");
                    }
                        
                    result.add("ALLOCATE(" + variableIterator.getName() + " % buffer(" +
                            entityIterator.getXCoordinate().getSize()
                            + " * " + entityIterator.getYCoordinate().getSize() + "))");
                    result.add(variableIterator.getName() +
                            " % two_dim  % successful_changes = 0");
                    result.add(variableIterator.getName() + " % buffer_index = -1");
                }
                
            }
        }
        
        //Calls for reading data and setting initial values
        result.add("");
        result.add("CALL read_data(" + callParameters + ")");        
        if (spatial != null)
            result.add("CALL set_spatial(spatial)");
        result.add("CALL set_initial_values(" + callParameters + ")");
        result.add("CALL set_functional(" + callParameters + ")");        
        
        
        result.add("");
        result.add("OPEN(UNIT=15, FILE=output_filename, STATUS='REPLACE', IOSTAT=status)");
        result.add("IF (status /= 0) THEN");
        result.add("WRITE (*, *) 'Could not open ', output_filename, ' for writing. Exiting.'");
        result.add("STOP");
        result.add("END IF");
        result.add("");
        result.add("OPEN(UNIT=14, FILE=summary_filename, STATUS='REPLACE', IOSTAT=status)");
        result.add("IF (status /= 0) THEN");
        result.add("WRITE (*, *) 'Could not open ', summary_filename, ' for writing. Exiting.'");
        result.add("STOP");
        result.add("END IF");
        
        result.add("");
        result.add("progress = 0");
        // generates a subroutine call
        
        if (model.getUpdateStrategy().equals("random")){
             // not implemented
            System.out.println("The random update strategy is not implemented.");
            throw new InvalidModelException("The random update strategy is not implemented.");
            
            //result.add("CALL update_one(" + callParameters + ", parameters_achieved)");
            
        }else if(model.getUpdateStrategy().equals("sequential")){
            //All variables are updated at the same time
            result.add("DO i=1, " + model.getIterations());
            result.add("CALL update_all(" + callParameters + ")");
            result.add("IF (i > burn_in .AND. MOD(i, thinning) == 0) THEN");
            result.add("CALL write_output("+callParameters + ")");
            result.add("END IF");
            result.add("");
            result.add("IF(i/("+ model.getIterations() + "/100) > progress) THEN");
            result.add("progress = i/("+ model.getIterations() + "/100)");
            result.add("WRITE(*, *) progress, '% of simulation complete'");
            result.add("END IF");
            
            result.add("END DO");
        }
        
        //Write summary here...
        result.add("");
        result.add("CALL write_summary(" + callParameters + ")");
        result.add("CALL write_last_values(" + callParameters + ")");
        
        result.add("CLOSE(15)");
        result.add("WRITE (*, *) 'Simulation completed successfully.'");
        result.add("");
        result.add("");
        result.add("CONTAINS");
        
        result.add("SUBROUTINE random_init()");
        result.add("! This subroutine initializes the default random number generator.");
        result.add("IMPLICIT NONE");
        result.add("EXTERNAL G05CBF, G05ZAF");
        result.add("INTEGER :: size, status");
        result.add("INTEGER, DIMENSION(8) :: time");
        result.add("INTEGER, DIMENSION(:), ALLOCATABLE :: seed");
        result.add("");
        result.add("CALL RANDOM_SEED(size=size)");
        result.add("ALLOCATE(seed(size), stat=status)");
        result.add("IF (status > 0) STOP 'Failed to allocate!'");
        result.add("CALL DATE_AND_TIME(values = time)");
        result.add("seed = 100*time(7) + time(8)/10");
        result.add("CALL G05ZAF('O')");
        result.add("CALL G05CBF(seed)");
        result.add("END SUBROUTINE random_init");
        result.add("");
        result.addAll(generateSetFunctional(callParameters, model));
        
        if (model.getUpdateStrategy().equals("random"))
            result.addAll(generateUpdateOne(callParameters, model));
        else if (model.getUpdateStrategy().equals("sequential"))
            result.addAll(generateUpdateAll(callParameters, model));
        
        result.add("END PROGRAM main");
        
        FortranWriter writer = new FortranWriter("main.f90");
        writer.write(result);
    }
    
    
    
    private static ArrayList<String> generateUpdateOne(String parameters, ComputationalModel
            model) {
         ArrayList<String> result = new ArrayList<String>();
        // Implemented iff we have time.
        return result;
    }
    
    private static ArrayList<String> generateUpdateAll(String parameters, ComputationalModel
            model)
            throws InvalidModelException, IllegalParametersException, MissingFunctionException {
        ArrayList<String> result = new ArrayList<String>();
        LinkedList<Variable> variables = model.getVariableList();
        LinkedList<Entity> entities = model.getEntityList();
        
        //Subroutine...
        result.add("SUBROUTINE update_all(" + parameters + ")");
        result.add("IMPLICIT NONE");
        
        //Variables...
        result.addAll(Generator.generateIntroduction());
     
        
        
        // !!! Introduce variables here !!!
        
        result.add("INTEGER, PARAMETER :: dp = SELECTED_REAL_KIND(8, 0)");
        result.add("INTEGER :: i, j, k, missing_index, prob_loop, ifail");
        result.add("INTEGER :: i2, j2, sum_int, neighbour, h");
        result.add("REAL (KIND=dp) :: p_acc, new_frequency, frequency");
        result.add("INTEGER :: int_transition");
        result.add("REAL (KIND=dp) :: transition_real, temp_real, sum_real");
        result.add("REAL (KIND=dp), SAVE, DIMENSION(1000) :: prob");
        result.add("INTEGER, SAVE :: prob_index = -1");
        
        // Distributions' freq functions need these
        result.add("EXTERNAL G01BKF");
        result.add("EXTERNAL G01BJF");
        result.add("EXTERNAL G01EEF");
        result.add("EXTERNAL G05FAF"); // for generating random number for acceptation
        
        
        result.add("IF (prob_index > SIZE(prob) .OR. prob_index < 1) THEN");
        result.add("CALL G05FAF(0.0_dp, 1.0_dp, SIZE(prob), prob)");
        result.add("prob_index = 1");
        result.add("END IF");
        
        
        String strategy = model.getUpdateStrategy();
        variables = model.getVariableList();
        
        /*
        // DEBUG
        System.out.println("GENERATOR DEBUG: printing global variables list:");
        for (Variable v : variables)
            System.out.println(v.getName());
        */
        for (Variable variableIterator : variables){
            if (!variableIterator.isFunctional()){
                result.add("! Updating " + variableIterator.getName());
                result.addAll(Acceptation.generateNewValueCode(variableIterator));
                result.addAll(Acceptation.generateNewValuesFunctionalCode(variableIterator));
                result.addAll(Acceptation.generateAcceptationFormula(variableIterator));
                result.addAll(Acceptation.generateAcceptationCode(variableIterator, strategy));
            }
        }
        
        
        for (Entity entityIterator : entities){
            variables = entityIterator.getVariableList();
            
            for (Variable variableIterator : variables){
                
                // DEBUG
                //System.out.println(variableIterator);
                
                if (variableIterator.isFunctional()){
                    continue;
                }
                
                if (variableIterator.isData() && variableIterator.getMissingValueCount() == 0){
                    continue;
                }
                result.add("! Updating " + variableIterator.getName());
                
                //We have an array...
                if (variableIterator.getEntity().isMatrix() == false){
                    //Variable is not data
                    if (!variableIterator.isData()) {
                        result.add("DO i = 1, " + variableIterator.getEntity().getSize());
                        result.addAll(Acceptation.generateNewValueCode(variableIterator));
                        result.addAll(Acceptation.generateNewValuesFunctionalCode(variableIterator));
                        result.addAll(Acceptation.generateAcceptationFormula(variableIterator));
                        result.addAll(Acceptation.generateAcceptationCode(variableIterator, strategy));
                        result.add("END DO");
                        
                        //Variable is data. Iterate through missing value array
                    } else {
                        result.add("DO missing_index = 1, " + variableIterator.getMissingValueCount());
                        result.add("i=" + variableIterator.getName() + " % one_dim_missing(missing_index)");
                        result.addAll(Acceptation.generateNewValueCode(variableIterator));
                        result.addAll(Acceptation.generateNewValuesFunctionalCode(variableIterator));
                        result.addAll(Acceptation.generateAcceptationFormula(variableIterator));
                        result.addAll(Acceptation.generateAcceptationCode(variableIterator, strategy));
                        result.add("END DO");
                        
                    }
                    
                    //We have a matrix
                } else {
                    //Variable is not data.
                    if (!variableIterator.isData()){
                        result.add("DO i=1, " + variableIterator.getEntity().getYCoordinate().getSize());
                        result.add("DO j=1, " + variableIterator.getEntity().getXCoordinate().getSize());
                        
                        result.addAll(Acceptation.generateNewValueCode(variableIterator));
                        result.addAll(Acceptation.generateNewValuesFunctionalCode(variableIterator));
                        result.addAll(Acceptation.generateAcceptationFormula(variableIterator));
                        result.addAll(Acceptation.generateAcceptationCode(variableIterator, strategy));
                        
                        result.add("END DO");
                        result.add("END DO");
                        
                        //Variable is data. Iterate through missing value matrix.
                    }else{
                        result.add("DO missing_index=1, " + variableIterator.getMissingValueCount());
                        result.add("i=" + variableIterator.getName() + " % two_dim_missing(missing_index, 1)");
                        result.add("j=" + variableIterator.getName() + " % two_dim_missing(missing_index, 2)");
                        result.addAll(Acceptation.generateNewValueCode(variableIterator));
                        result.addAll(Acceptation.generateNewValuesFunctionalCode(variableIterator));
                        result.addAll(Acceptation.generateAcceptationFormula(variableIterator));
                        result.addAll(Acceptation.generateAcceptationCode(variableIterator, strategy));
                        result.add("END DO");
                    }
                }
            }
        }
        
        result.add("END SUBROUTINE update_all");
        return result;
    }
    
    
    private static ArrayList<String> generateSetFunctional(String parameters, ComputationalModel
            model) throws InvalidModelException{
        
        ArrayList<String> result = new ArrayList<String>();
        LinkedList<Variable> variables = model.getVariableList();
        LinkedList<Entity> entities = model.getEntityList();
        
        result.add("");
        result.add("SUBROUTINE set_functional(" + parameters + ")");
        result.add("! This subroutine calculates initial values for functional parameters.");
        result.add("IMPLICIT NONE");
        
        //Variables...
        for (Variable variable : variables) {
            if (variable.isInteger())
                result.add("TYPE(variables_int), INTENT(INOUT) :: " + variable.getName());
            else
                result.add("TYPE(variables_real), INTENT(INOUT) :: " + variable.getName());
        }
        
        for (Entity entity : entities) {
            variables = entity.getVariableList();
            for (Variable variable : variables){
                if (variable.isInteger())
                    result.add("TYPE(variables_int), INTENT(INOUT) :: " + variable.getName());
                else
                    result.add("TYPE(variables_real), INTENT(INOUT) :: " + variable.getName());
            }
        }
        result.add("");
        result.add("INTEGER :: sum_int, k, i2, j2, neighbour");
        result.add("REAL (KIND=SELECTED_REAL_KIND(8, 0)) :: sum_real");
        result.add("");
        
        for (Entity entity : entities) {
            variables = entity.getVariableList();
            
            for (Variable variable : variables) {
                if (variable.isFunctional() && variable.isSpatial()){
                    result.addAll(correctLooping(variable, true));
                    result.addAll(setSpatialStartValue(variable)); // update spatials first
                    result.addAll(correctLooping(variable, false));
                    result.add("");
                }
            }
        }
        
        
        variables = model.getVariableList();
        
        for (Variable variable : variables) {
            if (variable.isSpatial()) // then the others
                continue;
            if (variable.isFunctional()){
                result.addAll(correctLooping(variable, true));
                result.addAll(setFunctionalStartValue(variable));
                result.addAll(correctLooping(variable, false));
                result.add("");
            }
        }
        
        for (Entity entity : entities) {
            variables = entity.getVariableList();
            for (Variable variable : variables){
                if (variable.isSpatial())
                    continue;
                if (variable.isFunctional()){
                    result.addAll(correctLooping(variable, true));
                    result.addAll(setFunctionalStartValue(variable));
                    result.addAll(correctLooping(variable, false));
                    result.add("");
                }
            }
        }
        
        
        result.add("END SUBROUTINE set_functional");
        result.add("");
        return result;
    }
    
    private static ArrayList<String> setSpatialStartValue(Variable var) throws InvalidModelException{
        
        //System.out.println("GENERATOR DEBUG: setSpatialStartValue for " + var.getName());
        //System.out.println("Details: " + var);

        ArrayList<String> result = new ArrayList<String>();
        
        String[] equation = var.getEquation().getEquation();
        String finalEquation = "";
        Variable par = var.getEquation().getParameters()[0];
        String parname = par.getName();
        String parIndexing = "";
        
        
        if (!var.getEntity().isMatrix()) {
            if (equation[0].startsWith("COUNT")){
                finalEquation = "spatial(i2, 1)";
                result.add(var.getName() + " % one_dim(i2) % value = " + finalEquation);
            } else{
                // calculate sum
                
                
                if (par.isInteger())
                    result.add("sum_int = 0");
                else
                    result.add("sum_real = 0");
                result.add("! 1 to number of neighbours");
                result.add("DO k = 1, spatial(i2, 1)");
                
                result.add("neighbour = spatial(i2, k + 1)");
                if (par.isInteger())
                    result.add("sum_int = sum_int + "+ parname + correctIndexing(par,
                            "neighbour") +" % value");
                else
                    result.add("sum_real = sum_real + "+ parname + correctIndexing(par,
                            "neighbour") +" % value");
                
                result.add("END DO");
                if (par.isInteger())
                    result.add(var.getName() + " % one_dim(i2) % value = sum_int");
                else
                    result.add(var.getName() + " % one_dim(i2) % value = sum_real");
                
            }
            
        } else{
            String indexing = "";
            if (var.getEntity().getYCoordinate().isSpatial())
                indexing = "i2";
            else
                indexing = "j2";
            
            if (equation[0].startsWith("COUNT")){
                finalEquation = "spatial("+ indexing +", 1)";
                result.add(var.getName() + " % two_dim(i2, j2) % value = " + finalEquation);
            } else{
                if (par.isInteger())
                    result.add("sum_int = 0");
                else
                    result.add("sum_real = 0");
                result.add("! 1 to number of neighbours");
                result.add("DO k = 1, spatial("+ indexing +", 1)");
                
                result.add("neighbour = spatial("+ indexing +", k + 1)");
                if (par.isInteger())
                    result.add("sum_int = sum_int + "+ parname + correctIndexing(par, "neighbour")
                    +" % value");
                else
                    result.add("sum_real = sum_real + "+ parname + correctIndexing(par, "neighbour")
                    +" % value");
                result.add("END DO");
                if (par.isInteger())
                    result.add(var.getName() + " % two_dim(i2, j2) % value = sum_int");
                else
                    result.add(var.getName() + " % two_dim(i2, j2) % value = sum_real");
                // calculate sum
                
            }
            
        }
        result.add("");
        return result;
    }
    private static ArrayList<String> setFunctionalStartValue(Variable var){
        ArrayList<String> result = new ArrayList<String>();
        
        String[] equation = var.getEquation().getEquation();
        /* the equation in parts: for example:
        3 * alpha + EXP(beta+gamma) ->
        3 *
        alpha
        + EXP(
        beta
        +
        gamma
        )
         */
        
        Variable[] parameters = var.getEquation().getParameters();
        
        for (int i=0; i<parameters.length; i++) {
            // dealing with Variable parameters[i]
            if (parameters[i].getEntity() == null) { // global
                for (int j=0; j<equation.length; j++) {
                    if (equation[j].equals(parameters[i].getName())) {
                        equation[j] += " % one_dim(1) % value";
                    }
                } // end looping the equation
            } else if (parameters[i].getEntity().isMatrix() == false) { // one-dimensional
                if (parameters[i].getEntity().equals(var.getEntity().getXCoordinate())) {
                    for (int j=0; j<equation.length; j++) {
                        if (equation[j].equals(parameters[i].getName())) {
                            equation[j] += " % one_dim(j2) % value";
                        }
                    } // end looping the equation
                } else { // either equals YCoordinate or both belong to the same entity 
                    for (int j=0; j<equation.length; j++) {
                        if (equation[j].equals(parameters[i].getName())) {
                            equation[j] += " % one_dim(i2) % value";
                            
                        }
                    } // end looping the equation
                }
                
            } else { // two-dimensional
                for (int j=0; j<equation.length; j++) {
                    if (equation[j].equals(parameters[i].getName())) {
                        equation[j] += " % two_dim(i2, j2) % value";
                        
                    }
                } // end looping the equation
            }
        } // end looping the parameters[]
        
        String finalEquation = "";
        for (int k=0; k<equation.length; k++) {
            finalEquation = finalEquation + equation[k];
        }
        
        if (var.getEntity() == null) {
            result.add(var.getName() + " % one_dim(1) % value = " + finalEquation);
        } else if (var.getEntity().isMatrix() == false) {
            result.add(var.getName() + " % one_dim(i2) % value = " + finalEquation);
        } else
            result.add(var.getName() + " % two_dim(i2, j2) % value = " + finalEquation);
        
        return result;
    }
    
    private static ArrayList<String> correctLooping(Variable var, boolean start){
        ArrayList<String> result = new ArrayList<String>();
        if (var.getEntity() == null)
            result.add("");
        else if (!var.getEntity().isMatrix()){
            if (start)
                result.add("DO i2 = 1, "+var.getEntity().getSize());
            else
                result.add("END DO");
        } else{
            if (start){
                result.add("DO i2 = 1, "+ var.getEntity().getYCoordinate().getSize());
                result.add("DO j2 = 1, "+ var.getEntity().getXCoordinate().getSize());
            } else{
                result.add("END DO");
                result.add("END DO");
            }
        }
        
        return result;
    }
    
    private static String correctIndexing(Variable par, String spatial)
    throws InvalidModelException{
        String parIndexing = null;
        if (par.getEntity() == null)
            throw new InvalidModelException("A spatial variable is affected by a global!");
        if (!par.getEntity().isMatrix()){
            parIndexing = " % one_dim ("+spatial+")";
        } else{
            if (par.getEntity().getYCoordinate().isSpatial())
                parIndexing = " % two_dim ("+spatial+", j2)";
            else
                parIndexing = " % two_dim (i2, "+spatial+")";
        }
        
        return parIndexing;
    }
}
